﻿using Franca;
using Franca.Gui;
using Shards;
using Shards.Environment;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Gui {
    public class WorldPreviewImage : Widget {
        enum ShowBorderType {
            None, Chunk, Tile
        }

        private int zoom = 0;
        private Point zoomPos;
        private ShowBorderType borders;

        public WorldPreviewImage(IWidget parent)
            : base(parent) {
            ClipContents = true;
            CanFocus = true;
        }

        protected override bool CaptureKey(KeyCaptureEvent e) {
            var key = KeyboardButton.KeyFromKeyCode(e.KeyCode);
            return key == Key.Left || key == Key.Right || key == Key.Up || key == Key.Down || key == Key.B || base.CaptureKey(e);
        }

        protected virtual void ProcessInput() {
            if (Zoomed) { //HasFocus) {
                var buffer = Input.Handler.TextBuffer;
                while (buffer.Count > 0) {
                    var k = buffer.Dequeue();
                    HandleKeyPress(k);
                }
            }
        }

        protected virtual void HandleKeyPress(KeyPress k) {
            if (zoom > 0) {
                var offset = zoom == 1 ? Chunk.Size : (zoom == 2 ? Chunk.Size / 2 : (zoom == 3 ? Chunk.Size / 4 : 1));
                switch (k.Key) {
                    case Key.Left:
                        zoomPos = new Point(zoomPos.X - offset, zoomPos.Y);
                        break;
                    case Key.Right:
                        zoomPos = new Point(zoomPos.X + offset, zoomPos.Y);
                        break;
                    case Key.Up:
                        zoomPos = new Point(zoomPos.X, zoomPos.Y - offset);
                        break;
                    case Key.Down:
                        zoomPos = new Point(zoomPos.X, zoomPos.Y + offset);
                        break;
                }
            }

            if (k.Key == Key.B) {
                var maxValue = Calc.Max(Enum.GetValues(typeof(ShowBorderType)).Cast<int>().ToArray()) + 1;
                borders = (ShowBorderType)Calc.Wrap((int)borders + 1, maxValue);
            }
        }

        protected override void MouseDown(GuiEvent e) {
            base.MouseDown(e);
            if (e.Button.Type == MouseButtons.Left) {
                if (zoom == 0) {
                    if (Game.Context.WorldPreview != null && Game.Context.WorldPreview.Progress.Done) {
                        var world = Game.Context.WorldPreview.World;
                        zoom = 1;
                        zoomPos = new Point(
                            (int)((e.X / (double)Width) * world.ChunkWidth * Chunk.Size),
                            (int)((e.Y / (double)Height) * world.ChunkHeight * Chunk.Size)
                        );
                    }
                }
                else {
                    zoom++;
                    if (zoom > 4) {
                        zoom = 4;
                    }
                }
            }
            else if (e.Button.Type == MouseButtons.Right) {
                zoom--;
                if (zoom < 0) {
                    zoom = 0;
                }
            }
        }

        public override void Layout(IGuiContext context) {
            base.Layout(context);
            //Width = 512;
            //Height = 512;
            ProcessInput();
        }

        public override void Render(IGuiContext context) {
            base.Render(context);
            var shards = context as IShardsGame;
            var batch = Manager.Batch;

            if (shards != null && shards.WorldPreview != null && shards.WorldPreview.Texture != null) {
                batch.Begin(filter: Franca.TextureFilter.Nearest);
                batch.Draw(shards.WorldPreview.Texture, DrawBounds, Color.White);
                batch.End();

                if (zoom > 0) {
                    var scale = zoom == 1 ? 0.25 : (zoom == 2 ? 0.5 : (zoom == 3 ? 1.0 : 2.0));
                    var tileSize = Calc.Round(32 * scale);
                    var world = shards.WorldPreview.World;
                    var renderer = shards.Renderer;
                    var size = Shards.Environment.Chunk.Size;
                    var centerX = DrawX + (Width - (size * tileSize)) / 2;
                    var centerY = DrawY + (Height - (size * tileSize)) / 2;
                    var range = Calc.Ceiling(Width / (size * tileSize));

                    batch.Begin(filter: zoom >= 3 ? TextureFilter.Nearest : TextureFilter.Default);
                    for (int x = -range; x <= range; x++) {
                        for (int y = -range; y <= range; y++) {
                            var chunk = world.GetChunk(zoomPos.X / Chunk.Size + x, zoomPos.Y / Chunk.Size + y);
                            if (chunk.BaseTiles == null) {
                                world.GenerateChunk(chunk.X, chunk.Y).Run(null);
                            }

                            var offsetX = zoomPos.X % Chunk.Size;
                            var offsetY = zoomPos.Y % Chunk.Size;

                            var chunkDrawX = centerX + size * tileSize * x - offsetX * tileSize;
                            var chunkDrawY = centerY + size * tileSize * y - offsetY * tileSize;

                            for (int tileX = 0; tileX < size; tileX++) {
                                for (int tileY = 0; tileY < size; tileY++) {
                                    renderer.DrawTile(batch, chunk.X * Chunk.Size + tileX, chunk.Y * Chunk.Size + tileY,
                                        new RectangleF(
                                            chunkDrawX + tileX * tileSize,
                                            chunkDrawY + tileY * tileSize,
                                            tileSize,
                                            tileSize
                                        ), Color.White);
                                }
                            }

                            foreach (var entity in (from e in chunk.Entities orderby e.X ascending, e.Y ascending select e)) {
                                renderer.DrawEntity(batch, entity, new Vector2(chunkDrawX + (entity.X % Chunk.Size) * tileSize, chunkDrawY + (entity.Y % Chunk.Size) * tileSize), scale, Color.White);
                            }

                            if (borders > ShowBorderType.None) {
                                if (borders == ShowBorderType.Tile) {
                                    for (int tileX = 0; tileX < size; tileX++) {
                                        for (int tileY = 0; tileY < size; tileY++) {
                                            batch.DrawRectangle(
                                                new RectangleF(chunkDrawX + (tileX + 1) * tileSize - 1, chunkDrawY + tileY * tileSize, 1, tileSize),
                                                Color.Black * 0.25);
                                            batch.DrawRectangle(
                                                new RectangleF(chunkDrawX + tileX * tileSize, chunkDrawY + (tileY + 1) * tileSize - 1, tileSize, 1), 
                                                Color.Black * 0.25);
                                        }
                                    }
                                }

                                batch.DrawRectangle(
                                    new RectangleF(chunkDrawX + size * tileSize - 1, chunkDrawY, 1, size * tileSize), 
                                    Color.Magenta * 0.25);
                                batch.DrawRectangle(
                                    new RectangleF(chunkDrawX, chunkDrawY + size * tileSize - 1, size * tileSize, 1), 
                                    Color.Magenta * 0.25);
                            }
                        }
                    }

                    batch.End();
                }
            }
            else {
                batch.Begin(filter: Franca.TextureFilter.Nearest);
                batch.DrawRectangle(DrawBounds, Color.Black * 0.5);
                batch.End();
            }
        }

        public bool Zoomed { get { return zoom > 0; } set { zoom = 0; } }
        public int Zoom { get { return zoom; } }
    }
}
